#include <ESP8266WiFi.h>
#include<LittleFS.h>


#define FILESYSTEM LittleFS
//#define FILESYSTEM SPIFFS


void dirRoot(){
    FILESYSTEM.begin();
    Dir dir = FILESYSTEM.openDir("/");
    printf("list root directory \"/\":\r\n");
    while (dir.next())
        printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());
    FILESYSTEM.end();
}


typedef class cDataBase{

    #define PREAMBLE (0x55A55AAA)
    // a complete record unit in a file:
    // [4-byte preamble]+[4-byte name length]+[4-byte name checksum]+[4-byte record length]+[n-byte name]+[m-byte record contents]=16+m+n.
    // payload is 16kbytes/1000records.

    typedef struct __attribute__((__packed__)) sRecord{
        unsigned nlen; // name length.
        unsigned ncs; // name checksum.
        unsigned rlen; // record length, name+id+data.
    } sRecord;

    void setup_filesystem(){
        FILESYSTEM.begin();
        static int x=1;
        if (x){
            x=0;
            {
                Dir dir = FILESYSTEM.openDir("/");
                if (!dir.next()){
                    FILESYSTEM.format();
                    printf("\r\n\r\nfilesystem formatted.\r\n");
                }
            }
            FSInfo fs_info;
            if (FILESYSTEM.info(fs_info)){
                printf(\
                    "--------\r\nThe FileSystem:\r\nentire space %d bytes\r\n%d bytes used\r\n"\
                    "free space %d bytes\r\nblock size %d bytes\r\npage size %d blocks\r\n--------\r\n",\
                    fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes, fs_info.blockSize, fs_info.pageSize);
            }
            Dir dir = FILESYSTEM.openDir("/");
            printf("list root directory \"/\":\r\n");
            while (dir.next())
                printf("%s    %s\r\n", dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());
        }
    };

    unsigned myFind(const void *str, unsigned sz){ // only for either char or unsigned char. 0/not found. others/the found position.
        if (!fh) return 0;
        unsigned i=0, j=0, w=sz*2, sum=0, csum=0, start_p=0, end_p=sz;
        unsigned char u[w+sz];
        int v=0;

        if (fh.readBytes((char*)u, sz)!=sz) return 0;
        for (; i<sz; i++){
            sum+=((const unsigned char*)str)[i];
            csum+=u[i];
        }

        while (1){
            if (sum==csum){
                for (i=0, j=0; i<sz && (((const unsigned char*)str)[i]==u[start_p+i-j]); i++) if (start_p+i-j==sz+w-1) j=start_p+i+1;
                if (i==sz){
                    fh.seek(-v, SeekCur);
                    return fh.position();
                }
            }
            if (!v){
                if (end_p==sz) v=fh.readBytes((char*)(u+end_p), w);
                else if (end_p==w+sz) v=fh.readBytes((char*)u, w);
                else {v=fh.readBytes((char*)(u+w), sz); v+=fh.readBytes((char*)u, sz);}
                if (!v) return v;
            }
            if (start_p==w+sz) start_p=0;
            if (end_p==w+sz) end_p=0;
            csum+=u[end_p++];
            csum-=u[start_p++];
            v--;
        }
    };

    const unsigned preamble=PREAMBLE;
    const char *db_name;
    File fh;

    public:

    cDataBase(): db_name(0){;};
    cDataBase(const char *dbn): db_name(dbn){;};
    void operator=(const char *dbn){db_name=dbn;};
    ~cDataBase(){this->Close();};

    void Open(){
        if (!db_name) return;
        setup_filesystem();
        fh=FILESYSTEM.open(db_name, "a+");
    };

    void Close(){
        if (fh) fh.close();
        db_name=0;
        FILESYSTEM.end();
    };

    void DumpAll(int f=0){ // 0: ascii; 1: hex.
        char u[8];
        int v;
        fh.seek(0, SeekSet);
        while ((v=fh.readBytes(u, 8))==8){
            printf("[");
            for (int i=0; i<8; i++)
                if (f) printf("%02X ", u[i]);
                else {
                    if (u[i]<' ' || u[i]>'~') u[i]='.';
                    printf("%2c ", u[i]);
                }
            printf("%s\r\n", "]");
        }
        printf("[");
        for (int i=0; i<v; i++)
            if (f) printf("%02X ", u[i]);
            else {
                if (u[i]<' ' || u[i]>'~') u[i]='.';
                printf("%2c ", u[i]);
            }
        printf("%s\r\n", "]");
    };

    template<typename T>
    void WriteWords(const char *name, int id, T data){
        // for elementary types, id whether unique or not maintained by user.
        // name is for the db to search purpose; whether unique or not maintained by user.

        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(data);

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((unsigned char*)&data, sizeof(data));
    };

    template<typename T>
    void WriteBytes(const char *name, int id, const T *buffer, unsigned length){
        sRecord u;

        for (u.nlen=0, u.ncs=0; name[u.nlen]; u.ncs+=name[u.nlen], u.nlen++);
        u.rlen=u.nlen+sizeof(id)+sizeof(T)*length; // note, not sizeof(buffer).

        fh.write((const unsigned char*)&preamble, sizeof(preamble));
        fh.write((const unsigned char*)&u, sizeof(u));
        fh.write((const unsigned char*)name, u.nlen);
        fh.write((unsigned char*)&id, sizeof(id));
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
    };

    template<typename T>
    void WriteBytesCont(const T *buffer, unsigned length){
        fh.write((const unsigned char*)buffer, sizeof(T)*length);
    };

    void FindStart(){ // a brand new start to find; from file ptr position 0.
        fh.flush();
        fh.seek(0, SeekSet);
    };

    unsigned FindNext(const char *name, int id=-1){ // find start from the current file position. file pos is moving up with FindNext.
    // return the fpos points to the begin of data(id+data contents) when id==-1. (and the fpos is where the current fptr pointing to).
    // return the fpos points to the begin of data contents when id!=-1. (and the fpos is where the current fptr pointing to).
        unsigned sum=0, i=0, len=0;
        sRecord u;
        for (; name[len]; len++) sum+=name[len];
        while (myFind(&preamble, sizeof(preamble))){
            if (fh.readBytes((char*)&u, sizeof(u))==sizeof(u)){
                if ((len==u.nlen) && (sum==u.ncs)){
                    char n[u.nlen];
                    if (fh.readBytes(n, u.nlen)==u.nlen){
                        for (i=0; i<u.nlen && (name[i]==n[i]); i++);
                        if (i==u.nlen){
                            if (id!=-1){
                                unsigned rid;
                                if (fh.readBytes((char*)&rid, sizeof(rid))==sizeof(rid)){
                                    if (id!=rid) fh.seek(u.rlen-u.nlen-sizeof(rid), SeekCur);
                                    else return fh.position(); // found data
                                }
                                else {/*fh.seek(0, SeekEnd); */return 0;}
                            }
                            else return fh.position(); // found id+data
                        }
                        else fh.seek(u.rlen-u.nlen, SeekCur);
                    }
                    else {/*fh.seek(0, SeekEnd); */return 0;}
                }
                else fh.seek(u.rlen, SeekCur);
            }
            else {/*fh.seek(0, SeekEnd); */return 0;}
        }
        /*fh.seek(0, SeekEnd);*/
        return 0;
    };

    template<typename T>
    bool ReadWords(T &data, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWords(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, int &id, const char *name){ // read both the data and id according to name. Next means start from current pos.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadWordsNext(T &data, const char *name, int id=-1){ // read the data only according to name or more with the id when id!=-1.
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)&data, sizeof(data))==sizeof(data)) return true;
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        FindStart();
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytes(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        FindStart();
        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, int &id, const char *name){ // read both the data and id according to name.
        if (FindNext(name)){
            if (fh.readBytes((char*)&id, sizeof(id))==sizeof(id)){
                if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
                return false;
            }
            return false;
        }
        return false;
    };

    template<typename T>
    bool ReadBytesNext(T *buffer, unsigned length, const char *name, int id=-1){
    // read the data only according to name or more with the id when id!=-1.

        if (FindNext(name, id)){
            if (id==-1) fh.seek(sizeof(id), SeekCur);
            if (fh.readBytes((char*)buffer, sizeof(T)*length)==sizeof(T)*length) return true; // note, not sizeof(buffer).
            return false;
        }
        return false;
    };

    template<typename T>
    void ReadBytesCont(T *buffer, unsigned length){
        fh.readBytes((char*)buffer, sizeof(T)*length);
    };

} cDataBase;


void setup(){
    Serial.begin(115200);
    delay(5000);

    double THE_PI=3.1415926f;
    cDataBase s;
    s="/database";
    s.Open();

    s.WriteWords("abcd", 4567, 1111.2f);
    s.WriteWords("defg", 7654, 6666);
    s.WriteBytes("1234", 0, "abcdefghijklm", 13);
    s.FindStart();
    printf("[%d %d %d]\r\n", s.FindNext("abcd"), s.FindNext("defg"), s.FindNext("1234"));
    s.DumpAll(1);

    s.WriteWords("doubleTHE_PI", 0, THE_PI);
    s.WriteWords("abcd", 4567, 1111.2f);
    s.WriteWords("fg", 7654, 6666);
    s.WriteBytes("dcba", 2, "abcdefghijklm", 13);
    s.WriteWords("dcba", 4567, 100000.f*100000.f*100000.f);
    s.WriteWords("gd", 1111, 6666);
    s.WriteBytes("eeeeee", 2, "abcdghijklm", 11);

    float L1, L4;
    int I1, I2, I3, I4, I5, I6;
    char C3[14]={0}, C6[12]={0};

    if (s.ReadWords(L4, "dcba", 4567)) printf("dcba[L4 %f]\r\n", L4);
    if (s.ReadBytesNext(C3, 13, I3, "dcba")) printf("dcba[I3a %d, %s]\r\n", I3, C3);

    L4=I4=I3=0;
    if (s.ReadBytes(C3, 13, I3, "dcba")) printf("dcba[I3b %d, %s]\r\n", I3, C3);
    if (s.ReadWordsNext(L4, I4, "dcba")) printf("dcba[L4 %d %f]\r\n", I4, L4);
    if (s.ReadBytesNext(C6, 11, "eeeeee", 1)) printf("eeeeee[C6a %s]\r\n", C6);
    if (s.ReadBytes(C6, 11, "eeeeee", 2)) printf("eeeeee[C6b %s]\r\n", C6);

    if (s.ReadWords(L1, I1, "dcba")) printf("abcd1[%d, %f]\r\n", I1, L1);
    if (s.ReadWords(L1, I1, "abcd")) printf("abcd2[%d, %f]\r\n", I1, L1);

    s.DumpAll(0);

    THE_PI=0;
    if (s.ReadWords(THE_PI, "doubleTHE_PI")) printf("\r\nvariable THE_PI[%f]\r\n", THE_PI);

    s.Close();
    dirRoot();
}

void loop(){
}
